local WagoAnalytics = LibStub("WagoAnalytics"):Register("9rN4XjND")
local SharedMedia = LibStub("LibSharedMedia-3.0");
local AddonName = "Jarspeedo"
local frame_cache = {}
local DebugMode = false;

local instanceRequirements;
local currCombatState = false;
local currInstance;
local currInstanceName;
local currTimerStartTime = 0;
local totalDisplayTime;
local totalTime;
local currSplits;
local compact = false;
local playerName = UnitName('player')
local partialRequirements = {};

if Splits_Configs == nil then
    Splits_Configs = {
        ["ShowTimesToBeat"] = true
    }
end

UI = nil

local TimesToBeat = {}

function AcquireFrame(parent, show)
    local frame = tremove(frame_cache) or CreateFrame("Frame")
    frame:SetParent(parent)
    if show then
        frame:Show()
    end
    return frame
end

function ReleaseFrame(frame)
    frame:Hide()
    frame:SetParent(nil)
    frame:ClearAllPoints()
    tinsert(frame_cache, frame)
end

function GetKeySortTable(tbl)
    keys = {}
    for key, _ in pairs(tbl) do
        table.insert(keys, key)
    end
    return keys;
end


function Split (inputstr, sep)
    if sep == nil then
            sep = "%s"
    end
    local t={}
    for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
            table.insert(t, str)
    end
    return t
end

function compareRelationshipWeights(a,b)
    return a[2] > b[2]
end

function sortAscending(a,b)
    return a > b
 end

 function sortRequirementIdAscending(a,b)
    return a.requirementId > b.requirementId
 end

 function sortTimestampAscending(a,b)
    return a.timestamp > b.timestamp
 end

function sortDescending(a,b)
   return a < b
end

function ReverseTable(t)
    local reversedTable = {}
    local itemCount = #t
    for k, v in ipairs(t) do
        reversedTable[itemCount+1 - k] = v
    end
    return reversedTable
end

function ReleaseLine(frame)
    if frame.left then
        frame.left:ClearAllPoints();
        frame.left = nil
    end
    if frame.right then
        frame.right:ClearAllPoints();
        frame.right = nil
    end
    if frame.center then
        frame.center:ClearAllPoints();
        frame.center = nil
    end

    frame:Hide()
    frame:SetParent(nil)
    frame:ClearAllPoints()
    tinsert(frame_cache, frame)
end

j_main_frame = AcquireFrame()

j_main_frame:RegisterEvent("CHAT_MSG_ADDON");
j_main_frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
j_main_frame:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT");
j_main_frame:RegisterEvent("ENCOUNTER_START");
j_main_frame:RegisterEvent("ENCOUNTER_END");
j_main_frame:RegisterEvent("UPDATE_INSTANCE_INFO");
j_main_frame:RegisterEvent("PLAYER_REGEN_DISABLED")
j_main_frame:RegisterEvent("PLAYER_REGEN_ENABLED")

SLASH_JARSPEEDOSHOWSPLITS1 = "/splits";

function ResetSplits(currInstance)
    ClearSplits()
    TimesToBeat = nil
    combatTimerTextFrame:SetText(nil)
    deathCountTextFrame:SetText(nil)
    timerTextFrame:SetText(nil)
    instanceRequirements = nil;
    currRequirements = nil;
    completedRequirements = nil;
    pendingRequirements = nil;
    if Requirements[currInstance] then
        for k,v in pairs(Requirements[currInstance]["Requirements"]) do
            v.count = 0;
        end
    end
    LoadRequirements(currInstance)
    RefreshUI()
end


function SlashCmdList.JARSPEEDOSHOWSPLITS(message, editbox)
    local args = Split(message, " ");
    if args[1] == "reset" then
        if args[2] ~= nil then
            local instance = MapShorthandInstanceNameToInstanceName(args[2])
            if Requirements[instance] ~= nil then
                currInstance = instance;
            else
                print("invalid usage: EXAMPLE: /splits reset tk")
            end
        end

        ResetSplits(currInstance)
        return
    elseif args[1] == "resetpbs" then
        if args[2] ~= nil then
            local instance = MapShorthandInstanceNameToInstanceName(args[2])
            if Requirements[instance] ~= nil then
                TimesToBeat = nil;
                Splits_Configs["TimesToBeat"][instance] = nil;
                RefreshUI()
            else
                print("invalid instance: EXAMPLE: /splits resetpbs tk")
            end
        else
            print("invalid usage: EXAMPLE: /splits resetpbs tk")
        end
    elseif args[1] ~= nil then
        local shorthand = MapShorthandInstanceNameToInstanceName(message)
        if shorthand then
            currInstance = shorthand
        elseif Requirements[message] then
            currInstance = message
        else
            return
        end
        pendingRequirementsOrdered = nil
        LoadRequirements(currInstance)
        RefreshUI()
        UI:Show();
        return;
    else
        if UI:IsShown() then
            UI:Hide();
        else
            RefreshUI();
            UI:Show();
        end
    end
end


    -- The minimum number of seconds between each update
local ONUPDATE_INTERVAL = 0.2

-- The number of seconds since the last update
local TimeSinceLastUpdate = 0
j_main_frame:SetScript("OnUpdate", function(self, elapsed)
    TimeSinceLastUpdate = TimeSinceLastUpdate + elapsed
    if TimeSinceLastUpdate >= ONUPDATE_INTERVAL then
        TimeSinceLastUpdate = 0
        
        if Splits_Configs[playerName] ~= nil and Splits_Configs[playerName][currInstance] and Splits_Configs[playerName][currInstance]["RunInProgress"] then
            local t = time();
            totalTime = t - currTimerStartTime
            totalDisplayTime = DisplayTime(totalTime);
            Splits_Configs[playerName][currInstance]["TotalDisplayTime"] = totalDisplayTime
            timerTextFrame:SetText(totalDisplayTime)
            
            if currCombatState and combatTimestamp then
                if currCombatTimer == nil then
                    combatTimestamp = t
                end
                currCombatTimer = (t - combatTimestamp)
                if Requirements[currInstance] and currCombatTimer > 0 then
                    combatTimerTextFrame:SetText("Combat: ("..DisplayTime(Splits_Configs[playerName][currInstance]["CombatTimeRunningTotal"] + currCombatTimer)..")")
                end
            end
        end
    end
end)

function SetTimesToBeat(sourceTable)
    local temp = sourceTable
    TimesToBeat = {}

    local kst = GetKeySortTable(temp)
    table.sort(kst, function(l,r) return temp[l].requirementId < temp[r].requirementId end)
    for k,v in ipairs(kst) do
        TimesToBeat[k] =  temp[v]
    end
end

function LoadRequirements(instanceId, lockoutId)
    if Requirements[instanceId] ~= nil then
        --if there's already a lockout lets load that lockout data
        if lockoutId ~= nil and Splits_Configs[playerName] ~= nil and Splits_Configs[playerName][instanceId] ~= nil then
            Splits_Configs[playerName][instanceId]["lockoutId"] = lockoutId
            instanceRequirements = Splits_Configs[playerName][instanceId];
        elseif Splits_Configs[playerName] ~= nil and Splits_Configs[playerName][instanceId] ~= nil then
            instanceRequirements = Splits_Configs[playerName][instanceId];
        else
            instanceRequirements = Requirements[instanceId];
        end

        if instanceRequirements["CompletedRequirementsCount"] == 0 then
            currRequirements = Requirements[instanceId]["Requirements"]
        else
            currRequirements = instanceRequirements["Requirements"]
        end
        
        if Splits_Configs["TimesToBeat"] ~= nil and Splits_Configs["TimesToBeat"][instanceId] ~= nil then
            SetTimesToBeat(Splits_Configs["TimesToBeat"][instanceId]["Splits"])
        else
            TimesToBeat = nil
        end

        if Splits_Configs[playerName] == nil then
            Splits_Configs[playerName] = {}
        end

        if Splits_Configs[playerName][instanceId] == nil then
            Splits_Configs[playerName][instanceId] = {
                ["CompletedRequirementsCount"] = 0,
                ["PendingRequirementsCount"] = #currRequirements,
                ["CombatTimeRunningTotal"] = 0
            }
        end

        if Splits_Configs[playerName][instanceId]["CurrTimerStartTime"] ~= nil then
            currTimerStartTime = Splits_Configs[playerName][instanceId]["CurrTimerStartTime"]
        else
            currTimerStartTime = 0
        end

        if Splits_Configs[playerName][instanceId]["TotalDisplayTime"] ~= nil then
            timerTextFrame:SetText(Splits_Configs[playerName][instanceId]["TotalDisplayTime"])
        else
            timerTextFrame:SetText(nil)
        end

        if Splits_Configs[playerName][instanceId]["Deaths"] ~= nil then
            deathCountTextFrame:SetText("Deaths: "..Splits_Configs[playerName][instanceId]["Deaths"])
        else
            deathCountTextFrame:SetText("Deaths: 0")
        end

        if Splits_Configs[playerName][instanceId]["CombatTimeRunningTotal"] ~= nil and Splits_Configs[playerName][instanceId]["CombatTimeRunningTotal"] > 0 then
            CombatTimeRunningTotal = Splits_Configs[playerName][instanceId]["CombatTimeRunningTotal"]
            combatTimerTextFrame:SetText("Combat: ("..DisplayTime(Splits_Configs[playerName][instanceId]["CombatTimeRunningTotal"])..")")
        else
            combatTimerTextFrame:SetText(nil)
        end

        if instanceRequirements["CompletedRequirements"] ~= nil then
            completedRequirements = instanceRequirements["CompletedRequirements"];
        else 
            completedRequirements = {}
        end

        if Splits_Configs[playerName][instanceId]["PartialRequirements"] ~= nil then
            partialRequirements = Splits_Configs[playerName][instanceId]["PartialRequirements"];
        else
            partialRequirements = {}
        end

        if instanceRequirements["PendingRequirements"] ~= nil then
            pendingRequirements = instanceRequirements["PendingRequirements"];

            for _, cv in pairs(completedRequirements) do

                if Requirements[instanceId]["Relationships"][cv.requirementId] ~= nil then 
                    for k,v in pairs(Requirements[instanceId]["Relationships"][cv.requirementId]) do
                        if pendingRequirements[k] ~= nil  then
                            pendingRequirements[k] = pendingRequirements[k] + v
                        end
                    end
                end
            end

            for pk, pv in pairs(partialRequirements) do

                if Requirements[instanceId]["Relationships"][pk] ~= nil then 
                    for k,v in pairs(Requirements[instanceId]["Relationships"][pk]) do
                        if pendingRequirements[k] ~= nil  then
                            pendingRequirements[k] = pendingRequirements[k] + v
                        end
                    end
                end
            end
            
            pendingRequirementsOrdered = {}
            for k,v in pairs(pendingRequirements) do
                if k ~= nil and v ~= nil then
                    table.insert(pendingRequirementsOrdered, {k, v})
                end
            end
            table.sort(pendingRequirementsOrdered, compareRelationshipWeights)
        else
            BuildPendingRequirements(currRequirements)
        end
    end
end

function BuildPendingRequirements(r)
    pendingRequirements = {};
    for k,v in ipairs(r) do
        pendingRequirements[k] = 0
    end
    Splits_Configs[playerName][currInstance]["PendingRequirementsCount"] = #pendingRequirements
end

local function OnMoveStart(self)
	if IsShiftKeyDown() then
		self:StartMoving()
	elseif IsAltKeyDown() then
		self:StartSizing("BOTTOMRIGHT")
	end
end

local function StopMovingOrSizing(self)
    Splits_Configs["PositionX"], Splits_Configs["PositionY"]  = self:GetCenter();
    Splits_Configs["FrameHeight"] = self:GetHeight();
    Splits_Configs["FrameWidth"] = self:GetWidth();

    self:StopMovingOrSizing();
end

function AddBorder(frame, color)
    local r,g,b = MapColorToColorObject(color)
    frame.left=frame:CreateTexture(nil,"BORDER")
    frame.left:SetPoint("BOTTOMLEFT",frame,"BOTTOMLEFT",-1,-1)
    frame.left:SetPoint("TOPRIGHT",frame,"TOPLEFT",0,1)
    frame.left:SetColorTexture(r,g,b)
    frame.right=frame:CreateTexture(nil,"BORDER")
    frame.right:SetPoint("BOTTOMLEFT",frame,"BOTTOMRIGHT",0,-1)
    frame.right:SetPoint("TOPRIGHT",frame,"TOPRIGHT",1,1)
    frame.right:SetColorTexture(r,g,b)
    frame.top=frame:CreateTexture(nil,"BORDER")
    frame.top:SetPoint("BOTTOMLEFT",frame,"TOPLEFT",-1,0)
    frame.top:SetPoint("TOPRIGHT",frame,"TOPRIGHT",1,1)
    frame.top:SetColorTexture(r,g,b)
    frame.bottom=frame:CreateTexture(nil,"BORDER")
    frame.bottom:SetPoint("BOTTOMLEFT",frame,"BOTTOMLEFT",-1,0)
    frame.bottom:SetPoint("TOPRIGHT",frame,"BOTTOMRIGHT",1,-1)
    frame.bottom:SetColorTexture(r,g,b)
end

function UpdateBorderColor(frame, color)
    local r,g,b = MapColorToColorObject(color)
    frame.left:SetColorTexture(r,g,b);
    frame.right:SetColorTexture(r,g,b);
    frame.top:SetColorTexture(r,g,b);
    frame.bottom:SetColorTexture(r,g,b);
end

function LoadAddon()
    if UI ~= nil then return end

    local font;   

    if Splits_Configs["MainFont"] then
        font = Splits_Configs["MainFont"]
    else
        font = "Fonts\\ARIALN.ttf"
    end
    WagoAnalytics:IncrementCounter("Jarspeedo_Loaded")

    --ui
    UI = AcquireFrame(UIParent);
    UI:SetScript("OnShow", function(self)     
        minimizeButton:Show()
        encountersButton:Show()
        interfacePanelButton:Show()
        UITimer:Show()        
    end) 
    UI:SetScript("OnHide", function(self) 
        minimizeButton:Hide()
        encountersButton:Hide()
        interfacePanelButton:Hide()
        UITimer:Hide()
    end) 

    UI.bg = UI:CreateTexture(nil, "BACKGROUND")
    UI.bg:SetAllPoints(true)

    if Splits_Configs["BgColor"] then
        UI.bg:SetColorTexture(Splits_Configs["BgColor"].r, Splits_Configs["BgColor"].g, Splits_Configs["BgColor"].b, Splits_Configs["BgColor"].a)
    else
        UI.bg:SetColorTexture(0.18,0.2,0.16,0.35)
    end

    UI:SetClipsChildren(true);
    UI:EnableMouse(true);
    UI:SetMovable(true)
    UI:SetResizable(true)    
    UI:RegisterForDrag("LeftButton");
    UI:SetScript("OnDragStart", OnMoveStart)
    UI:SetScript("OnDragStop", StopMovingOrSizing);
    UI:SetMinResize(128,15)
    UI:SetMaxResize(2000,1000)

    if Splits_Configs["PositionX"] then
        UI:SetWidth(Splits_Configs["FrameWidth"]);
        UI:SetHeight(Splits_Configs["FrameHeight"]);
        UI:SetPoint("CENTER", UIParent, "BOTTOMLEFT", Splits_Configs["PositionX"], Splits_Configs["PositionY"])
    else
        UI:SetWidth(300);
        UI:SetHeight(300);
        UI:SetPoint("CENTER", 0, 0)
    end

    minimizeButton = AcquireFrame(UIParent)
    minimizeButton:SetPoint("BOTTOMLEFT", UI, "TOPLEFT",0,1)
    minimizeButton:SetSize(10,10)
    AddBorder(minimizeButton, "gray")

    minimizeButtonText = minimizeButton:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    minimizeButtonText:SetFont("Fonts\\ARIALN.ttf", 10)
    minimizeButtonText:SetPoint("BOTTOMLEFT", minimizeButton.bottom, "BOTTOMLEFT",minimizeButton.bottom:GetWidth()/4,1)
    minimizeButtonText:SetTextColor(MapColorToColorObject("gray"))
    minimizeButtonText:SetText("M")


    minimizeButton.text = minimizeButtonText;

    minimizeButton:SetScript("OnEnter", 
        function(self, arg1)
            self.text:SetTextColor(MapColorToColorObject("yellow"))
            UpdateBorderColor(self, "yellow")
        end)
    minimizeButton:SetScript("OnLeave", 
        function(self, arg1)
            if compact then 
                self.text:SetTextColor(MapColorToColorObject("green"))
                UpdateBorderColor(self, "green")
            else
                self.text:SetTextColor(MapColorToColorObject("gray"))
                UpdateBorderColor(self, "gray")
            end
        end)
    minimizeButton:SetScript("OnMouseDown", 
        function(self, arg1)
            
            compact = not compact

            if compact then 
                self.text:SetTextColor(MapColorToColorObject("green"))
                UpdateBorderColor(self, "green")
            else
                self.text:SetTextColor(MapColorToColorObject("gray"))
                UpdateBorderColor(self, "gray")
            end
            RefreshUI();
        end)

    encountersButton = AcquireFrame(UIParent)
    encountersButton:SetPoint("LEFT", minimizeButton, "RIGHT",5,0)
    encountersButton:SetSize(10,10)
    AddBorder(encountersButton, "gray")

    encountersButtonText = encountersButton:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    encountersButtonText:SetFont("Fonts\\ARIALN.ttf", 10)
    encountersButtonText:SetPoint("BOTTOMLEFT", encountersButton.bottom, "BOTTOMLEFT",encountersButton.bottom:GetWidth()/4,1)
    encountersButtonText:SetTextColor(MapColorToColorObject("gray"))
    encountersButtonText:SetText("B")
    encountersButton.text = encountersButtonText;

    encountersButton:SetScript("OnEnter", 
        function(self, arg1)
            self.text:SetTextColor(MapColorToColorObject("yellow"))
            UpdateBorderColor(self, "yellow")
        end)
        encountersButton:SetScript("OnLeave", 
        function(self, arg1)
            if onlyEncounters then 
                self.text:SetTextColor(MapColorToColorObject("green"))
                UpdateBorderColor(self, "green")
            else
                self.text:SetTextColor(MapColorToColorObject("gray"))
                UpdateBorderColor(self, "gray")
            end
        end)
        encountersButton:SetScript("OnMouseDown", 
        function(self, arg1)
            
            onlyEncounters = not onlyEncounters

            if onlyEncounters then 
                self.text:SetTextColor(MapColorToColorObject("green"))
                UpdateBorderColor(self, "green")
            else
                self.text:SetTextColor(MapColorToColorObject("gray"))
                UpdateBorderColor(self, "gray")
            end
            RefreshUI();
        end)

    interfacePanelButton = AcquireFrame(UIParent)
    interfacePanelButton:SetPoint("LEFT", encountersButton, "RIGHT",5,0)
    interfacePanelButton:SetSize(10,10)
    AddBorder(interfacePanelButton, "gray")

    interfacePanelButtonText = interfacePanelButton:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    interfacePanelButtonText:SetFont("Fonts\\ARIALN.ttf", 10)
    interfacePanelButtonText:SetPoint("BOTTOMLEFT", interfacePanelButton.bottom, "BOTTOMLEFT",interfacePanelButton.bottom:GetWidth()/4,1)
    interfacePanelButtonText:SetTextColor(MapColorToColorObject("gray"))
    interfacePanelButtonText:SetText("?")
    interfacePanelButton.text = interfacePanelButtonText;

    interfacePanelButton:SetScript("OnEnter", 
        function(self, arg1)
            self.text:SetTextColor(MapColorToColorObject("yellow"))
            UpdateBorderColor(self, "yellow")
        end)
    interfacePanelButton:SetScript("OnLeave", 
        function(self, arg1)
            self.text:SetTextColor(MapColorToColorObject("gray"))
            UpdateBorderColor(self, "gray")
        end)
    interfacePanelButton:SetScript("OnMouseDown", 
        function(self, arg1)
            --blizzard bug makes the first call to this not hone into specific addon, leaves on general options page, wowpedia lists same bug as known and recommends double call
            if InterfaceOptionsCalledOnce then
                InterfaceOptionsFrame_OpenToCategory(AddonName)
            else
                InterfaceOptionsCalledOnce = true;
                InterfaceOptionsFrame_OpenToCategory(AddonName)
                InterfaceOptionsFrame_OpenToCategory(AddonName)
            end
            self.text:SetTextColor(MapColorToColorObject("gray"))
            UpdateBorderColor(self, "gray")
        end)


    UITimer = AcquireFrame(UIParent);
    timerTextFrame = UITimer:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    timerTextFrame:SetPoint("TOPRIGHT", UI, "BOTTOMRIGHT",0,0)
    timerTextFrame:SetFont(font, 24)
    timerTextFrame:SetTextColor(1,1,1,1)
    LoadComponents()
    UI:Hide()
end

function LoadComponents()

    if Splits_Configs["MainFont"] then
        font = Splits_Configs["MainFont"]
    else
        font = "Fonts\\ARIALN.ttf"
    end

    if Splits_Configs["DefaultFontSize"] then
        fontSize = Splits_Configs["DefaultFontSize"]
    else
        fontSize = 15
    end

    local componentYOffset = -2;

    if not combatTimerTextFrame then
        combatTimerTextFrame = UITimer:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    end
    
    if Splits_Configs["ShowCombatTimer"] == true and combatTimerTextFrame:GetText() then
        combatTimerTextFrame:SetPoint("TOPLEFT", UI, "BOTTOMLEFT",0,componentYOffset)
        combatTimerTextFrame:SetFont(font, fontSize)
        combatTimerTextFrame:SetTextColor(1,.5,.5,1)
        componentYOffset = componentYOffset - fontSize;
        combatTimerTextFrame:Show();
    else
        combatTimerTextFrame:Hide();
    end

    if not deathCountTextFrame then
        deathCountTextFrame = UITimer:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    end

    if Splits_Configs["ShowDeathCount"] == true then
        deathCountTextFrame:SetPoint("TOPLEFT", UI, "BOTTOMLEFT",0,componentYOffset)
        deathCountTextFrame:SetFont(font, fontSize)
        deathCountTextFrame:SetTextColor(1,.5,.5,1)
        componentYOffset = componentYOffset - fontSize;

        deathCountTextFrame:Show();
    else
        deathCountTextFrame:Hide();
    end

end

function RefreshUI()
    if not UI then return end;

    if UI.lines then
        ClearTable(UI.lines)
    end

    UI.lines = {}

    
    local index = 1
    local numCompletedRecordsToHide = 0;
    local numPendingRecordsToHide = 0;
    local fontSize = 15;

    if Splits_Configs["DefaultFontSize"] then
        fontSize = Splits_Configs["DefaultFontSize"]
    end
    
    if compact and not onlyEncounters then
        if Splits_Configs["NumCompletedRecordsToShow"] == nil then
            Splits_Configs["NumCompletedRecordsToShow"] = 3
        end

        if Splits_Configs["NumPendingRecordsToShow"] == nil then
            Splits_Configs["NumPendingRecordsToShow"] = 5
        end

        if currInstance then
                numCompletedRecordsToHide = Splits_Configs[playerName][currInstance]["CompletedRequirementsCount"] - Splits_Configs["NumCompletedRecordsToShow"]

            if numCompletedRecordsToHide < 0 then 
                numCompletedRecordsToHide = 0;
            end

            numPendingRecordsToShow = Splits_Configs["NumPendingRecordsToShow"]
        end
    end

    --complete
    local completeIndex = 0;

    if completedRequirements ~= nil then

        local centerTextMaxWidth = 0;
        local kst = GetKeySortTable(completedRequirements)
        table.sort(kst, function(l,r) return completedRequirements[l].timestamp < completedRequirements[r].timestamp end)
        for k,v in ipairs(kst) do
            k = completedRequirements[v].requirementId
            v = completedRequirements[v]
            if onlyEncounters then
                if Requirements[currInstance]["Encounters"][v.requirementId] then
                    centerTextMaxWidth = AddSplitLineItem(nil, v.requirementId, currRequirements[v.requirementId].count, currRequirements[v.requirementId].requirement, 0, (index-1 - numCompletedRecordsToHide) * -fontSize, "green", centerTextMaxWidth, fontSize)                     
                    index = index + 1;
                    completeIndex = completeIndex + 1;
                end
            elseif compact then
                if  index > numCompletedRecordsToHide then
                    if not onlyEncounters or onlyEncounters and Requirements[currInstance]["Encounters"][v.requirementId] then
                        centerTextMaxWidth = AddSplitLineItem(nil, v.requirementId, currRequirements[v.requirementId].count, currRequirements[v.requirementId].requirement, 0, (index-1 - numCompletedRecordsToHide) * -fontSize, "green", centerTextMaxWidth, fontSize)                     
                    end
                end
                index = index + 1;
                completeIndex = completeIndex + 1;
            else
                centerTextMaxWidth = AddSplitLineItem(nil, v.requirementId, currRequirements[v.requirementId].count, currRequirements[v.requirementId].requirement, 0, (index-1 - numCompletedRecordsToHide) * -fontSize, "green", centerTextMaxWidth, fontSize)                     
                index = index + 1;
                completeIndex = completeIndex + 1;
            end
        end
    end

    index = index - numCompletedRecordsToHide

    --partially complete
    local partialIndex = 0;

    if partialRequirements ~= nil then
        local centerTextMaxWidth = 0;
        for k,v in pairs(partialRequirements) do
            if  onlyEncounters then
                if onlyEncounters and Requirements[currInstance]["Encounters"][k] then
                    centerTextMaxWidth = AddSplitLineItem(nil, k, currRequirements[k].count, currRequirements[k].requirement, 0, (index-1) * -fontSize, "yellow", centerTextMaxWidth, fontSize)
                    index = index + 1;
                    partialIndex = partialIndex + 1;
                end
            else
                centerTextMaxWidth = AddSplitLineItem(nil, k, currRequirements[k].count, currRequirements[k].requirement, 0, (index-1) * -fontSize, "yellow", centerTextMaxWidth, fontSize)
                index = index + 1;
                partialIndex = partialIndex + 1;
            end
        end
    end

    --pending
    if pendingRequirements ~= nil then
        
        local centerTextMaxWidth = 0;
        local t

        if pendingRequirementsOrdered ~= nil then
            t = pendingRequirementsOrdered
        else
            t = pendingRequirements
        end

        for k,v in pairs(t) do
            if type(v) == 'table' then 
                k = v[1]
            end

            if onlyEncounters then 
                if onlyEncounters and Requirements[currInstance]["Encounters"][k] then
                    centerTextMaxWidth = AddSplitLineItem(nil, k, 0, currRequirements[k].requirement, 0, (index-1) * -fontSize, "red", centerTextMaxWidth, fontSize)
                    index = index + 1;
                end
            elseif compact then
                --if index + numPendingRecordsToHide < Splits_Configs[playerName][currInstance]["PendingRequirementsCount"] + completeIndex + partialIndex then
                if numPendingRecordsToShow > 0 then
                    centerTextMaxWidth = AddSplitLineItem(nil, k, 0, currRequirements[k].requirement, 0, (index-1) * -fontSize, "red", centerTextMaxWidth, fontSize)
                    numPendingRecordsToShow = numPendingRecordsToShow - 1;
                end
                index = index + 1;
            else
                centerTextMaxWidth = AddSplitLineItem(nil, k, 0, currRequirements[k].requirement, 0, (index-1) * -fontSize, "red", centerTextMaxWidth, fontSize)
                index = index + 1;
            end
        end
    end
    
    LoadComponents()
end

function ClearTable(t)
    for k,v in pairs (t) do
        ReleaseLine(v)
    end
end

function AddSplitLineItem(index, requirementId, curr, req, x, y, color, centerTextMaxWidth, fontSize)

    local newline = AcquireFrame(UI, true);
    local font;   

    if Splits_Configs["MainFont"] then
        font = Splits_Configs["MainFont"]
    else
        font = "Fonts\\ARIALN.ttf"
    end

    newline.left = newline:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    newline.left:SetFont(font, fontSize)
    newline.left:SetTextColor(MapColorToColorObject(color))

    local killRequirementString = "";
    --don't show 1/1 for single kill requirements, implied by default
    if not (curr == req == 1) then
        killRequirementString = " "..curr.."/"..req
    end

    --if we have a starttime for the split we can show the split completion time

    local leftText = L[currInstance][requirementId];

    if MapRequirementIdToEncounter(requirementId) == nil then
        leftText = leftText..killRequirementString;
    end

    if completedRequirements[requirementId] ~= nil and completedRequirements[requirementId].startTime ~= nil then
        newline.left:SetText(leftText.." ("..DisplayTime(completedRequirements[requirementId].timestamp - completedRequirements[requirementId].startTime)..")")
    else
        newline.left:SetText(leftText)
    end
    newline.left:SetPoint("TOPLEFT", UI, "TOPLEFT",x,y-2)


    if Splits_Configs["ShowTimesToBeat"] == true and TimesToBeat ~= nil and TimesToBeat[requirementId] then

        newline.right = newline:CreateFontString(nil, "OVERLAY", "GameFontNormal")
        newline.right:SetFont(font, fontSize)
        newline.right:SetTextColor(MapColorToColorObject(color))
        newline.right:SetText("("..DisplayTime(TimesToBeat[requirementId].timestamp)..")")
        newline.right:SetPoint("TOPRIGHT", UI, "TOPRIGHT",x,y-2)
    end

    local referenceframe = UI

    if newline.right then 
        referenceframe = newline.right;
        x = -75
        y = 0
    end

    newline.center = newline:CreateFontString(nil, "OVERLAY", "GameFontNormal")
    newline.center:SetFont(font, fontSize)
    newline.center:SetTextColor(MapColorToColorObject(color))

    if completedRequirements[requirementId] ~= nil then
        local centerText = DisplayTime(completedRequirements[requirementId].timestamp - currTimerStartTime)
        newline.center:SetText(centerText)
        local w = newline.center:GetWidth()

        if centerTextMaxWidth and w < centerTextMaxWidth then
            newline.center:SetWidth(centerTextMaxWidth)
        else
            centerTextMaxWidth = w;
        end

    else
        newline.center:SetText("--:--")
    end
    newline.center:SetPoint("TOPRIGHT", referenceframe, "TOPRIGHT", x, y-2)

    table.insert(UI.lines, newline);
    return centerTextMaxWidth;
end

j_main_frame:SetScript("OnEvent", function(self, event, ...)
    --if event == "CHAT_MSG_ADDON" and arg ~= nil then
     --   HandleAddonMessage(unpack(arg))    
    if event == "ENCOUNTER_START" then
        self:ENCOUNTER_START(...)
    elseif event == "ENCOUNTER_END" then
        self:ENCOUNTER_END(...)    
    elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
        self:COMBAT_LOG_EVENT_UNFILTERED(CombatLogGetCurrentEventInfo())
    elseif event == "UPDATE_INSTANCE_INFO" then
        self:UPDATE_INSTANCE_INFO()
    elseif event == "PLAYER_REGEN_DISABLED" then
        currCombatState = true;
        combatTimestamp = time();
    elseif Splits_Configs[playerName] ~= nil and Splits_Configs[playerName][currInstance] ~= nil and Splits_Configs[playerName][currInstance]["RunInProgress"] ~= nil and event == "PLAYER_REGEN_ENABLED" then

        if currCombatTimer == nil then 
            currCombatTimer = 0 
        end

        if currInstance ~= nil then        
            Splits_Configs[playerName][currInstance]["CombatTimeRunningTotal"] = Splits_Configs[playerName][currInstance]["CombatTimeRunningTotal"] + currCombatTimer
            currCombatTimer = nil;
            currCombatState = false;
        end
    end
end)

function j_main_frame:ENCOUNTER_START(...)
    local encounterId,encounterName = ...

    DebugLog("Encounter Start:", encounterId, encounterName)
    local requirementId = MapUnitToRequirementId(encounterId)
    if not requirementId then 
        return;
    end

    HandleEncounterEngagement(requirementId, time())
end

function j_main_frame:ENCOUNTER_END( ...)
    local encounterId,encounterName,a1,a2,success = ...

    if success == 0 then 
        return
    end

    local requirementId = MapUnitToRequirementId(encounterId)

    if not requirementId then 
        return;
    end
    
    if completedRequirements[requirementId] ~= nil then
        return
    end

    for i=1, GetNumSavedInstances() do
        local savedInstanceName, lockoutId = GetSavedInstanceInfo(i)
        if savedInstanceName == currInstanceName then
            Splits_Configs[playerName][currInstance]["lockoutId"] = lockoutId
        end
    end
    IncrementRequirementCompletedCount(requirementId, time())
end

function RemoveAndRebuildPendingRequirements(requirementId)
    pendingRequirements[requirementId] = nil;

    if Requirements[currInstance]["Relationships"][requirementId] == nil then
        return
    end

    for k,v in pairs(Requirements[currInstance]["Relationships"][requirementId]) do
        if pendingRequirements[k] ~= nil  then
            pendingRequirements[k] = pendingRequirements[k] + v
        end
    end

    pendingRequirementsOrdered = {}
    for k,v in pairs(pendingRequirements) do
        if k ~= nil and v ~= nil then
            table.insert(pendingRequirementsOrdered, {k, v})
        end
    end
    table.sort(pendingRequirementsOrdered, compareRelationshipWeights)
end

function StartNewRun(timestamp)
    ResetSplits(currInstance)
    currTimerStartTime = timestamp
    Splits_Configs[playerName][currInstance]["CurrTimerStartTime"] = currTimerStartTime
    Splits_Configs[playerName][currInstance]["RunInProgress"] = true;
end

function HandleEncounterEngagement(requirementId, timestamp)

    if Splits_Configs[playerName][currInstance] and Splits_Configs[playerName][currInstance]["RunInProgress"] == nil then
        if Requirements[currInstance]["ValidRequirementStartIds"][requirementId] == true then
            DebugLog("StartNewRun: HandleEncounterEngagement: requirementId:"..requirementId)
            StartNewRun(timestamp)
       end
    end

    if partialRequirements[requirementId] or completedRequirements[requirementId] then
        return;
    end

    DebugLog("HandleEncounterEngagement requirementId: "..requirementId)

    partialRequirements[requirementId] = {startTime = timestamp}

    RemoveAndRebuildPendingRequirements(requirementId)
    Splits_Configs[playerName][currInstance]["PendingRequirementsCount"] = Splits_Configs[playerName][currInstance]["PendingRequirementsCount"] - 1


    SaveSplits();
    RefreshUI();
end

function j_main_frame:UPDATE_INSTANCE_INFO(...)
    
    LoadAddon()
    instanceName, instanceType,_,_,arg3,arg2,arg1,instanceId = GetInstanceInfo()
    
    if Requirements[instanceId] == nil then 
        return
    end
    
    currInstance = instanceId
    currInstanceName = instanceName

    for i=1, GetNumSavedInstances() do
        local savedInstanceName, lockoutId = GetSavedInstanceInfo(i)
        if savedInstanceName == currInstanceName then
            LoadRequirements(currInstance, lockoutId)
            RefreshUI()
            return;
        end
    end

    LoadRequirements(currInstance)--, 92833647)
    RefreshUI()
end

function HandleRaidMemberDeath()
    if Splits_Configs[playerName][instanceId]["Deaths"] ~= nil then
        Splits_Configs[playerName][instanceId]["Deaths"] = Splits_Configs[playerName][instanceId]["Deaths"] + 1;
    else
        Splits_Configs[playerName][instanceId]["Deaths"] = 0;
    end
    deathCountTextFrame:SetText("Deaths: "..Splits_Configs[playerName][instanceId]["Deaths"])
end

function j_main_frame:COMBAT_LOG_EVENT_UNFILTERED(...)
    local timestamp, subevent, arg1, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellId, spellName, arg3, arg4, arg5, arg6, arg7 = ...

    if Requirements[currInstance] ~= nil then
        if subevent == "UNIT_DIED" then
            if UnitInRaid(destName) and Splits_Configs[playerName][currInstance]["RunInProgress"] then
                HandleRaidMemberDeath()
                return;
            else
                local creatureId = Split(destGUID, '-')[6]

                HandleUnitDeath(timestamp, creatureId)
                return;
            end
        end

        if subevent == "SPELL_CAST_SUCCESS" then
            local creatureId = Split(destGUID, '-')[6]

            if (not Requirements["BlacklistedAbilities"][spellId]) and UnitInRaid(sourceName) then
                HandleSpellCastSuccess(timestamp, creatureId)
            end
            return;
        end
    end
end

function HandleUnitDeath(timestamp, creatureId)   
    local requirementId = MapUnitToRequirementId(creatureId)

    if (not requirementId) or requirementId < 0 then 
        return;
    end

    if MapRequirementIdToEncounter(requirementId) ~= nil then   
        if completedRequirements[requirementId] ~= nil then
            return
        end
    end

    IncrementRequirementCompletedCount(requirementId, timestamp)
end

function HandleSpellCastSuccess(timestamp, creatureId)      

    local requirementId = MapUnitToRequirementId(creatureId)

    if not requirementId then 
        return;
    end

    HandleNonEncounterRequirementBegin(requirementId, timestamp)
end


function IncrementRequirementCompletedCount(requirementId, timeStamp)

    currRequirements[requirementId].count = currRequirements[requirementId].count + 1;

    DebugLog("IncrementRequirementCompletedCount requirementId: "..requirementId.." count: "..currRequirements[requirementId].count)

    if currRequirements[requirementId].count == currRequirements[requirementId].requirement then
        --completion finished, update UI
        HandleRequirementCompletion(requirementId, timeStamp)
    end

    RefreshUI();
end

function SaveSplits() 
    Splits_Configs[playerName][currInstance]["Requirements"] = currRequirements;
    Splits_Configs[playerName][currInstance]["CompletedRequirements"] = completedRequirements;
    Splits_Configs[playerName][currInstance]["PartialRequirements"] = partialRequirements;
    Splits_Configs[playerName][currInstance]["PendingRequirements"] = pendingRequirements;
end

function ClearSplits()
    pendingRequirementsOrdered = nil;
    partialRequirements = nil;
    if Splits_Configs[playerName][currInstance] then
        Splits_Configs[playerName][currInstance] = nil;
    end
end

function HandleRequirementCompletion(requirementId, timestamp)
    
    DebugLog("HandleRequirementCompletion requirementId: "..requirementId)

    if partialRequirements[requirementId] ~= nil then
        partialRequirements[requirementId].timestamp = timestamp
        partialRequirements[requirementId].requirementId = requirementId

        completedRequirements[requirementId] = partialRequirements[requirementId]

        partialRequirements[requirementId] = nil;
    else
        completedRequirements[requirementId] =  {timestamp = timestamp, requirementId = requirementId}
    end

    RemoveAndRebuildPendingRequirements(requirementId)

    Splits_Configs[playerName][currInstance]["CompletedRequirementsCount"]  = Splits_Configs[playerName][currInstance]["CompletedRequirementsCount"]  + 1;

    SaveSplits();

    if AreAllRequirementsSatisfied() == true then
        Splits_Configs[playerName][currInstance]["RunInProgress"] = nil;
        UpdateTimesToBeat(timestamp)
    end
end

function AreAllRequirementsSatisfied()
    for k,_ in pairs(partialRequirements) do
        if k ~= nil then return false end
    end

    for k,_ in pairs(pendingRequirements) do
        if k ~= nil then return false end
    end
    return true;
end

function UpdateTimesToBeat(finalTimeStamp)

    totalTime = finalTimeStamp - currTimerStartTime
    totalDisplayTime = DisplayTime(totalTime);
    Splits_Configs[playerName][currInstance]["TotalDisplayTime"] = totalDisplayTime
    timerTextFrame:SetText(totalDisplayTime)
    
    if Splits_Configs["TimesToBeat"] == nil then
        Splits_Configs["TimesToBeat"] = {}
    end

    if Splits_Configs["TimesToBeat"][currInstance] == nil  or (totalTime < Splits_Configs["TimesToBeat"][currInstance]["BestTime"]) then
        Splits_Configs["TimesToBeat"][currInstance] = {["BestTime"] = totalTime, ["Splits"] = {}}
    else 
        return;
    end

    
    DebugLog("UpdateTimesToBeat")

    for k,v in ipairs(completedRequirements) do        

        table.insert(Splits_Configs["TimesToBeat"][currInstance]["Splits"], 1,  {requirementId = v.requirementId, timestamp = v.timestamp - currTimerStartTime})
    end

    SetTimesToBeat(Splits_Configs["TimesToBeat"][currInstance]["Splits"])

end

function HandleNonEncounterRequirementBegin(requirementId, timestamp)
    --negative requirementId means valid start mob but not tied to a kill requirement
    if requirementId < 0 and (Splits_Configs[playerName][currInstance] ~= nil and Splits_Configs[playerName][currInstance]["RunInProgress"] == nil) then
        DebugLog("StartNewRun requirementId < 0: HandleNonEncounterRequirementBegin requirementId:"..requirementId)
        StartNewRun(timestamp)
        return;
    elseif requirementId < 0 then 
        return
    end

    if Splits_Configs[playerName][currInstance] ~= nil and Splits_Configs[playerName][currInstance]["RunInProgress"] == nil then
        if Requirements[currInstance]["ValidRequirementStartIds"][requirementId] == true then      
           DebugLog("StartNewRun: HandleNonEncounterRequirementBegin requirementId:"..requirementId)
           StartNewRun(timestamp)
        end
    end

    if pendingRequirements[requirementId] == nil or partialRequirements[requirementId] ~= nil or completedRequirements[requirementId] ~= nil then
        return;
    end

    DebugLog("HandleNonEncounterRequirementBegin requirementId:"..requirementId)

    partialRequirements[requirementId] = {startTime = timestamp}

    RemoveAndRebuildPendingRequirements(requirementId)

    Splits_Configs[playerName][currInstance]["PendingRequirementsCount"] = Splits_Configs[playerName][currInstance]["PendingRequirementsCount"] - 1


    SaveSplits();
    RefreshUI();
end

function MapUnitToRequirementId(unitId)
    if Requirements[currInstance] ~= nil and Requirements[currInstance]["ValidMobs"] then
        return Requirements[currInstance]["ValidMobs"][tonumber(unitId)];
    end
end

function MapRequirementIdToEncounter(requirementId)
    return Requirements[currInstance]["Encounters"][requirementId];
end

function MapShorthandInstanceNameToInstanceName(shorthand)
    return Requirements["ShorthandInstanceNameMap"][shorthand];
end

function MapColorToColorObject(color)
    if color == "green" then
        local c =  Splits_Configs["CompletedRequirementsColor"] 
        if c then
            return c.r, c.g, c.b, c.a
        else
            return .224, .863, .137, 1
        end
    elseif color == "yellow" then
        local c =  Splits_Configs["PartialRequirementsColor"] 
        if c then
            return c.r, c.g, c.b, c.a
        else
            return .714, .714, .137, 1    
        end
    --We're gonna call this red even though it defaults to the same as gray
    elseif color == "red" then
        local c =  Splits_Configs["PendingRequirementsColor"] 
        if c then
            return c.r, c.g, c.b, c.a
        else
            return .824, .824, .824, 1
        end
    elseif color == "gray" then
        return .824, .824, .824, 1
    end
end

function DisplayTime(time)
    local hours = floor(mod(time, 86400)/3600)
    local minutes = floor(mod(time,3600)/60)
    local seconds = floor(mod(time,60))
    if hours == 0 then    
        return format("%d:%02d",minutes,seconds)
    end
    return format("%d:%02d:%02d",hours,minutes,seconds)
end

function DebugLog(line)
    if DebugMode ~= true then
        return
    end
    if not Splits_Configs["Logs"] then
        local newtable = {}
        table.insert(newtable, time()..": "..line)
        Splits_Configs["Logs"] = newtable
        return
    end
    if currInstance then
        table.insert(Splits_Configs["Logs"], time().." | "..currInstance.." | "..line)
    else
        table.insert(Splits_Configs["Logs"], time().." | NONE | "..line)
    end
end

function BackgroundColorPickCallback(restore)
    local newR, newG, newB, newA;

    if restore then
     -- The user bailed, we extract the old color from the table created by ShowColorPicker.
     newR, newG, newB, newA = unpack(restore);
    else
     -- Something changed
     newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB();

     --no idea why this is backwards
     newA = 1 - newA;
    end
    
    if UI then
        UI.bg:SetColorTexture(newR, newG, newB, newA)
    end

    if panel.bgButton then
        panel.bgButton.bg:SetColorTexture(newR, newG, newB, newA)
    end

    Splits_Configs["BgColor"] = {r = newR, g = newG, b = newB, a = newA};
end

function CompletedRequirementsColorPickCallback(restore)
    local newR, newG, newB, newA;

    if restore then
     -- The user bailed, we extract the old color from the table created by ShowColorPicker.
     newR, newG, newB, newA = unpack(restore);
    else
     -- Something changed
     newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB();

     --no idea why this is backwards
     newA = 1 - newA;
    end
    
    Splits_Configs["CompletedRequirementsColor"] = {r = newR, g = newG, b = newB, a = newA};

    if panel.completedRequirementsColorButton then
        panel.completedRequirementsColorButton.bg:SetColorTexture(newR, newG, newB, newA)
    end

    if UI then
        RefreshUI();
    end
end

function PartialRequirementsColorPickCallback(restore)
    local newR, newG, newB, newA;

    if restore then
     -- The user bailed, we extract the old color from the table created by ShowColorPicker.
     newR, newG, newB, newA = unpack(restore);
    else
     -- Something changed
     newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB();

     --no idea why this is backwards
     newA = 1 - newA;
    end
    
    Splits_Configs["PartialRequirementsColor"] = {r = newR, g = newG, b = newB, a = newA};

    if panel.partialRequirementsColorButton then
        panel.partialRequirementsColorButton.bg:SetColorTexture(newR, newG, newB, newA)
    end

    if UI then
        RefreshUI();
    end
end

function PendingRequirementsColorPickCallback(restore)
    local newR, newG, newB, newA;

    if restore then
     -- The user bailed, we extract the old color from the table created by ShowColorPicker.
     newR, newG, newB, newA = unpack(restore);
    else
     -- Something changed
     newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB();

     --no idea why this is backwards
     newA = 1 - newA;
    end
    
    Splits_Configs["PendingRequirementsColor"] = {r = newR, g = newG, b = newB, a = newA};

    if panel.pendingRequirementsColorButton then
        panel.pendingRequirementsColorButton.bg:SetColorTexture(newR, newG, newB, newA)
    end

    if UI then
        RefreshUI();
    end
end

function UpdateMainFont(font)
    Splits_Configs["MainFont"] = font;
    local fontSize = 15;

    if Splits_Configs["DefaultFontSize"] then
        fontSize = Splits_Configs["DefaultFontSize"]
    end
    combatTimerTextFrame:SetFont(font, fontSize)
    deathCountTextFrame:SetFont(font, fontSize)
    timerTextFrame:SetFont(font, fontSize+9)
    RefreshUI();
end

function UpdateDefaultFontSize(fontSize)
    Splits_Configs["DefaultFontSize"] = fontSize;

    if Splits_Configs["MainFont"] then
        font = Splits_Configs["MainFont"]
    else
        font = "Fonts\\ARIALN.ttf"
    end

    combatTimerTextFrame:SetFont(font, fontSize)
    deathCountTextFrame:SetFont(font, fontSize)
    timerTextFrame:SetFont(font, fontSize+9)
    RefreshUI();
end